const sleep = require("sleep-promise");
const BDKAgentClient = require("../EndpointClients/BDKAgent");
const AFCMakeCallEndpointBase = require("./AFCMakeCallEndpointBase");
const powerSingleton = require("../Modules").PowerSingleton;

const ServerInfoGatherer = require("../../BDKAgentModules/ServerInfoGatherer").ServerInfoGathererSingleton;
const ServerCleanup = require("../../BDKAgentModules/ServerCleanup");
const ServerRegistering = require("../../BDKAgentModules/ServerRegistering");
const UsedServerBoot = require("../../BDKAgentModules/UsedServerBoot");
const CommandExecutor = require("../../BDKAgentModules/CommandExecutor");
const {Encryption} = require("encryption");

module.exports =
	class BDKAgent extends AFCMakeCallEndpointBase
	{
		constructor(agentIdentity)
		{
			super(
				/*strName*/ "BDKAgent",
				/*strPath*/ "/api-ws/BDKAgent",
				/*objReflection*/ {},
				/*classReverseCallsClient*/ BDKAgentClient
			);

			this._agentIdentity = agentIdentity;
			this._serverEncryption = new Encryption(this._agentIdentity.serverEncryptionRotatingKeys);
			this._objAFCPromises = {};
			this._bUpdatePrerequisiteFinished = false;
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForce = false
		 * @param {number} nTimeoutMilliseconds = 0
		 * @returns {string}
		 */
		async bdk_agent_reboot(incomingRequest, bForce, nTimeoutMilliseconds)
		{
			await powerSingleton.powerOff(/*bReboot*/ true, bForce, nTimeoutMilliseconds);

			return "Server will now reboot.";
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForce = false
		 * @param {number} nTimeoutMilliseconds = 0
		 * @returns {string}
		 */
		async bdk_agent_shutdown(incomingRequest, bForce, nTimeoutMilliseconds)
		{
			await powerSingleton.powerOff(/*bReboot*/ false, bForce, nTimeoutMilliseconds);

			return "Server will now shutdown.";
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {number} nSleepSeconds
		 * @returns {string}
		 */
		async bdk_agent_test(incomingRequest, nSleepSeconds)
		{
			const strStartTime = new Date().toISOString();
			await sleep(nSleepSeconds * 1000);

			return "Hello! Test start time: " + strStartTime + ". End time: " + (new Date().toISOString());
		}


		/**
		 * Collects various hardware information needed to update the newly created server in the database
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForceRegenerate
		 * @returns {Object}
		 */
		async bdk_agent_server_info_collect(incomingRequest, bForceRegenerate)
		{
			return await ServerInfoGatherer.gatherAllInfo(bForceRegenerate);
		}


		/**
		 * Collects general information needed to update the newly created server in the database
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForceRegenerate
		 * @returns {Object}
		 */
		async bdk_agent_server_general_info(incomingRequest, bForceRegenerate)
		{
			return await ServerInfoGatherer.gatherGeneralInfo(bForceRegenerate);
		}


		/**
		 * Collects cpu information needed to update the newly created server in the database
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForceRegenerate
		 * @returns {Object}
		 */
		async bdk_agent_server_cpu_info(incomingRequest, bForceRegenerate)
		{
			return await ServerInfoGatherer.gatherCPUInfo(bForceRegenerate);
		}


		/**
		 * Collects bios information needed to update the newly created server in the database
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForceRegenerate
		 * @returns {Object}
		 */
		async bdk_agent_server_bios_info(incomingRequest, bForceRegenerate)
		{
			return await ServerInfoGatherer.gatherBiosInfo(bForceRegenerate);
		}


		/**
		 * Collects disks information needed to update the newly created server in the database
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForceRegenerate
		 * @returns {Object}
		 */
		async bdk_agent_server_disks_info(incomingRequest, bForceRegenerate)
		{
			return await ServerInfoGatherer.gatherDiskInfo(bForceRegenerate);
		}


		/**
		 * Collects memory information needed to update the newly created server in the database
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForceRegenerate
		 * @returns {Object}
		 */
		async bdk_agent_server_memory_info(incomingRequest, bForceRegenerate)
		{
			return await ServerInfoGatherer.gatherMemoryInfo(bForceRegenerate);
		}


		/**
		 * Collects memory information needed to update the newly created server in the database
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForceRegenerate
		 * @returns {Object}
		 */
		async bdk_agent_server_network_info(incomingRequest, bForceRegenerate)
		{
			return await ServerInfoGatherer.gatherNetworkInterfaceInfo(bForceRegenerate);
		}


		/**
		 * Execute all required cleanup stages:
		 * - networkingIntelCleanup
		 * - serverInterfacesCleanup
		 * - IPMIUsersAndNetworkingCleanup
		 * - efibootmgrCleanup
		 * - diskCleanup
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Object} objServer
		 * @param {Array<Object>} arrServerInterfaces
		 * @param {Array} objIPOOB
		 * @param {Array} arrDNSServers
		 * @param {Array} arrNTPServers
		 * @returns {string}
		 */
		async bdk_agent_server_cleanup(incomingRequest, objServer, arrServerInterfaces, objIPOOB, arrDNSServers, arrNTPServers)
		{
			return await ServerCleanup.BDKCleanup.run(this._agentIdentity, objServer, arrServerInterfaces, objIPOOB, arrDNSServers, arrNTPServers);
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Object} objServer
		*/
		async bdk_agent_networking_intel_cleanup(incomingRequest, objServer)
		{
			return await ServerCleanup.BDKCleanup.networkingIntelCleanup(objServer);
		}

		/**
			* @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Array} arrServerInterfaces
		*/
		async bdk_agent_server_interfaces_cleanup(incomingRequest, arrServerInterfaces)
		{
			return await ServerCleanup.BDKCleanup.serverInterfacesCleanup(arrServerInterfaces);
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Object} objServer
		 * @param {Array} objIPOOB
		 * @param {Array} arrDNSServers
		 * @param {Array} arrNTPServers
		*/
		async bdk_agent_ipmi_users_and_networking_cleanup(incomingRequest, objServer, objIPOOB, arrDNSServers, arrNTPServers)
		{
			return await ServerCleanup.BDKCleanup.IPMIUsersAndNetworkingCleanup(this._agentIdentity, objServer, objIPOOB, arrDNSServers, arrNTPServers);
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Array} arrNICDetails
		*/
		async bdk_agent_efibootmgr_cleanup(incomingRequest, arrNICDetails)
		{
			return await ServerCleanup.BDKCleanup.efibootmgrCleanup(arrNICDetails);
		}

		async bdk_agent_disks_cleanup(incomingRequest, objServer)
		{
			return await ServerCleanup.BDKCleanup.diskCleanup(this._agentIdentity, objServer);
		}

		/**
	 	 * This function will set up IPMI for the server.
		 * Credentials and IP addresses should be created or allocated before calling this function.
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Object} [objConfig]
		 * @param {bool} [bOnlyEnableIPMIOverLan]
		 * @returns {Promise<number>}
		 */
		async bdk_agent_ipmi_setup(incomingRequest, objConfig, bOnlyEnableIPMIOverLan = false)
		{
			return await ServerRegistering.IPMISetup.run(objConfig, this._serverEncryption, bOnlyEnableIPMIOverLan);
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest 
		 * @param {Array<Object>} arrServerInterfaces 
		 * @param {Object} objInstance 
		 */
		async bdk_agent_hba_setup(incomingRequest, arrServerInterfaces, objInstance)
		{
			return await UsedServerBoot.HBASetup.run(this._agentIdentity, arrServerInterfaces, objInstance);
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest 
		 * @param {string} strCommand 
		 */
		async bdk_run_command_batch(incomingRequest, strCommand)
		{
			return await CommandExecutor.execCommandProcess(strCommand);
		}
		

		/**
		 * Exits the current process after 2 seconds using setTimeout().
		 * Should return before the setTimeout's callback executes. Two seconds should be enough time to return.
		 *
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * 
		 * @returns {string} 
		 */
		async bdk_agent_process_exit(incomingRequest)
		{
			setTimeout(
				() =>
				{
					try
					{
						console.log("Will process.exit(1) now.");
						process.exit(1);
					}
					catch (error)
					{
						console.error(error);
					}
				},
				2000
			);
			
			return "Will process.exit(1) in 2 seconds";
		}
	};
