// @ts-check

const CommandBase = require("./CommandBase");

class NetworkInterfaces extends CommandBase
{
	constructor(objConfig = {})
	{
		super(objConfig);
	}

	static get command()
	{
		return `
			IFACES=""
			for f in /sys/class/net/*
			do
				if [ $f == "/sys/class/net/lo" ]; then continue; fi
				ifindex=\`cat $f/ifindex\`
				ifindex=$((ifindex -1))
				mac=\`cat $f/address\`
				IFACES="\${IFACES}__BSI__INTERFACE \$ifindex \$mac\n"
				name=\`echo $f | sed "s/\\\/sys\\\/class\\\/net\\\///g\"\`
				speed=\`ethtool \${name} | grep Speed\`
				IFACES="\${IFACES}\${speed}\n"
				/usr/sbin/lldptool -L -i \${name} adminStatus=rxtx
				sleep 10s
				IFACES="\${IFACES}__START_LLDP_INFO"
				LLDP=\`/usr/sbin/lldptool -t -i \${name} -n | sed 's/'"\$(printf '\\015')"'//g'\`
				if [ -z "$LLDP" ]; then
					sleep 10s
					echo "Retried LLDP on an interface once"
					LLDP=\`/usr/sbin/lldptool -t -i \${name} -n | sed 's/'"\$(printf '\\015')"'//g'\`
					if [ -z "$LLDP" ]; then
						sleep 10s
						echo "Retried LLDP on an interface twice"
						LLDP=\`/usr/sbin/lldptool -t -i \${name} -n | sed 's/'"\$(printf '\\015')"'//g'\`
					fi
				fi
				IFACES="\${IFACES}\${LLDP}"
				IFACES="\${IFACES}__END_INTERFACE_INFO"
			done
		
			echo \${IFACES}
		`;
	}

	/**
	 * Parses the raw output from the command, must me runned after execCommand
	 * 
	 * @param {string} [strResult=null]
	 * @returns {Promise<Object>} _objParsedCommandOutput
	 */
	async _parseCommandOutput(strResult = null)
	{
		if(strResult)
		{
			this._strRawCommandOutput = strResult;
		}

		if(!this._strRawCommandOutput || this._strRawCommandOutput.trim() === "")
		{
			throw new Error(`Empty output string for command ${this.constructor.name}.`);
		}
		
		const objParsedResult = {};
		const arrNetworkRows = this._strRawCommandOutput.split("__END_INTERFACE_INFO");
		const objInterfaceInformation = {};

		objInterfaceInformation.interfaces = [];
		objParsedResult.server_network_total_capacity_mbps = 0;

		for(let nIndex = 0; nIndex < arrNetworkRows.length; nIndex++)
		{
			let strInterfaceRow = arrNetworkRows[nIndex];

			if(strInterfaceRow.trim().length === 0)
			{
				continue;
			}

			if(strInterfaceRow.includes("__BSI__INTERFACE"))
			{
				strInterfaceRow = strInterfaceRow.split("__BSI__INTERFACE")[1];

				const arrInformation = strInterfaceRow.split("__START_LLDP_INFO");
				objInterfaceInformation.interfaces[nIndex] = this._parseLLDPInformation(arrInformation[1]);

				const arrHeaderInformation = arrInformation[0].split("Speed: ");
				const arrBasicInformation = arrHeaderInformation[0].split(" ");

				objInterfaceInformation.interfaces[nIndex]["server_interface_index"] = parseInt(arrBasicInformation[1], 10) - 1;
				objInterfaceInformation.interfaces[nIndex]["server_interface_mac_address"] = arrBasicInformation[2].trim();

				const strSpeed = arrHeaderInformation[1];
				const arrSpeeds = strSpeed.split("Mb/s");
				
				let nCurrentNetworkCapacity = null;

				if(arrSpeeds.length === 2)
				{
					nCurrentNetworkCapacity = parseInt(arrSpeeds[0], 10);
				}
				else
				{
					nCurrentNetworkCapacity = 0;
				}

				objInterfaceInformation.interfaces[nIndex]["server_interface_capacity_mbps"] = nCurrentNetworkCapacity;

				objParsedResult.server_network_total_capacity_mbps += nCurrentNetworkCapacity;
			}
		};
		
		this._objParsedCommandOutput = {...objParsedResult, ...objInterfaceInformation};
		return this._objParsedCommandOutput;
	}

	/**
	 * Extract the necessary LLDP information and put it in an object
	 * 
	 * @param {string} strLLDPInformation
	 * @returns {Object} objLLDPInformation
	 */
	_parseLLDPInformation(strLLDPInformation)
	{
		const objTempLLDPInformation = {};
		const objLLDPInformation = {};

		objLLDPInformation["network_equipment_interface_lldp_information"] = strLLDPInformation;

		strLLDPInformation = strLLDPInformation.replace(/\n\t|\n /, "\t");
		// const arrLLDPRows = strLLDPInformation.split("\n");

		const objRequiredLLDPInformation = {
			"Chassis ID TLV": "network_equipment_interface_mac_address",
			"Port ID TLV": "switch_port_id",
			"Port Description TLV": "switch_port_description",
			"System Name TLV": "switch_hostname",
			"System Description TLV": "network_equipment_description",
			"Port VLAN ID TLV": "switch_vlan_id",
			"VLAN Name TLV": "switch_vlan_name"
		};

		for(let strInfoName in objRequiredLLDPInformation)
		{
			if(strLLDPInformation.includes(strInfoName))
			{
				const strPropertyName = objRequiredLLDPInformation[strInfoName];
				let strInfo = strLLDPInformation.split(strInfoName)[1];
		
				switch (strInfoName) 
				{
					case "Chassis ID TLV":
						strInfo = strInfo.split(" Port ID TLV")[0];
						break;
					case "Port ID TLV":
						strInfo = strInfo.split(" Time to Live TLV")[0];
						break;
					case "Port Description TLV":
						strInfo = strInfo.split(" System Name TLV")[0];
						break;
					case "System Name TLV":
						strInfo = strInfo.split(" System Description TLV")[0];
						console.log(strInfo);
						break;
					case "System Description TLV":
						strInfo = strInfo.split(" System Capabilities TLV")[0];
						break;
					case "Port VLAN ID TLV":
						strInfo = strInfo.split(" Link Aggregation TLV")[0];
						break;
					case "VLAN Name TLV":
						strInfo = strInfo.split(" MAC/PHY Configuration Status TLV")[0];
						break;
					default:
						break;
				}
		
				if(strInfo.includes(": "))
				{
					strInfo = strInfo.split(": ")[1];
				}
		
				objLLDPInformation[strPropertyName] = strInfo.trim();
			}
		}

		// arrLLDPRows.forEach((strLLDPRow) =>
		// {
		// 	let arrLLDPRow = strLLDPRow.split("\t");

		// 	if(arrLLDPRow.length >= 2)
		// 	{
		// 		objTempLLDPInformation[arrLLDPRow[0]] = arrLLDPRow.shift() && arrLLDPRow.join("");
		// 	}
		// });

		// for(let strLLDPKey in objTempLLDPInformation)
		// {
		// 	if(objRequiredLLDPInformation[strLLDPKey])
		// 	{
		// 		let strLLDPValue = objTempLLDPInformation[strLLDPKey];
		// 		if(strLLDPValue.includes(": "))
		// 		{
		// 			strLLDPValue = strLLDPValue.split(": ")[1];
		// 		}

		// 		objLLDPInformation[objRequiredLLDPInformation[strLLDPKey]] = strLLDPValue;
		// 	}
		// }

		return objLLDPInformation;
	}

	get parsedOutput()
	{
		return this._objParsedCommandOutput;
	}

	static get osPlatform()
	{
		return "linux";
	}

	async test()
	{
		
	}
}

module.exports = NetworkInterfaces;
